# -*- encoding: utf-8 -*-

import os
import environ
import logging
from pathlib import Path

from django.utils import timezone

log = logging.getLogger(__name__)

VERSION = "0.1.10"

# BASE_DIR is the root of the project, all paths should be constructed from it using pathlib
BASE_DIR = Path(__file__).parent.parent

# Create an environment and read variables from .env file
env = environ.Env(DEBUG=(bool, True))
environ.Env.read_env(BASE_DIR / ".env")

required_env_vars = (
    "BOT_TOKEN", "DISCORD_KEY", "DISCORD_SECRET", "DISCORD_REDIRECT_URL",
    "DISCORD_OAUTH2_URL", "SUPERUSER_IDS"
)

for var in required_env_vars:
    if not env(var, default=None):
        log.warn("Required environment variable %s is not set, the application will fail!", var)

# SECURITY WARNING: This is sensitive data, keep secure!
SECRET_KEY = env('SECRET_KEY', default="unsecure-default-secret-key")

# SECURITY WARNING: Must be 'False' in production!
DEBUG = env('DEBUG')

# Hosts and Origins that the server host must be within.
ALLOWED_HOSTS = ["localhost", "127.0.0.1", env("HOST", default="127.0.0.1")]
CSRF_TRUSTED_ORIGINS = ["http://localhost", "http://127.0.0.1", "https://" + env("HOST", default="127.0.0.1")]

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    "rest_framework.authtoken",
    "django_filters",
    'apps.api',
    'apps.home',
    'apps.authentication',
    "apps.health",
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'core.urls'
APPEND_SLASH = True

LOGIN_URL = "/login/"
LOGIN_REDIRECT_URL = "/"
LOGOUT_REDIRECT_URL = "/"

AUTH_USER_MODEL = "authentication.DiscordUser"

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR / "apps/templates"],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'core.wsgi.application'


# Database
# https://docs.djangoproject.com/en/5.0/ref/settings/#databases

DB_ENGINE = env("DB_ENGINE", default=None)

if not DB_ENGINE:
    db_data = {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3"
    }
else:
    db_data = {
        "ENGINE": env("DB_ENGINE"),
        "NAME": env("DB_NAME"),
        "USER": env("DB_USER"),
        "PASSWORD": env("DB_PASS"),
        "HOST": env("DB_HOST"),
        "PORT": env("DB_PORT")
    }

DATABASES = { "default": db_data }

# Password validation
# https://docs.djangoproject.com/en/5.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    { "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator" },
    { "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator" },
    { "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator" },
    { "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator" }
]

AUTHENTICATION_BACKENDS = [
    # "django.contrib.auth.backends.ModelBackend",
    "apps.authentication.backends.DiscordAuthenticationBackend"
]

# Discord Related Settings

BOT_TOKEN = env("BOT_TOKEN", default=None)
DISCORD_KEY = env("DISCORD_KEY", default=None)
DISCORD_SECRET = env("DISCORD_SECRET", default=None)
DISCORD_SCOPES = ["identify", "guilds"]
DISCORD_CODE_EXCHANGE_REQUEST = {
    "headers": {"Content-Type": "application/x-www-form-urlencoded"},
    "data": {
        "client_id": DISCORD_KEY,
        "client_secret": DISCORD_SECRET,
        "grant_type": "authorization_code",
        "redirect_uri": env("DISCORD_REDIRECT_URL", default=None),
        "scope": " ".join(DISCORD_SCOPES),
        "code": ""
    }
}
DISCORD_REFRESH_TOKEN_REQUEST = {
    "headers": {"Content-Type": "application/x-www-form-urlencoded"},
    "data": {
        "grant_type": "refresh_token",
        "refresh_token": ""
    }
}
DISCORD_API_URL = env("DISCORD_API_URL", default="https://discord.com/api/v10")
DISCORD_OAUTH2_URL = env("DISCORD_OAUTH2_URL", default=None)
SUPERUSER_IDS = env("SUPERUSER_IDS", default="").split(",")

# Logging
# https://docs.djangoproject.com/en/5.0/topics/logging/

LOGGING_DIR = BASE_DIR / "logs"
LOGGING_DIR.mkdir(exist_ok=True)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': LOGGING_DIR / f'{timezone.now()}.log',
            "formatter": "verbose",
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            "formatter": "verbose"
        },
        "timed_file": {
            "level": "DEBUG",
            "class": "logging.handlers.TimedRotatingFileHandler",
            "when": "D",
            "interval": 1,
            "backupCount": 3,
            "encoding": "UTF-8",
            "filename": LOGGING_DIR / "debug.log",
            "formatter": "verbose"
        }
    },
    'loggers': {
        "apps": {
            "handlers": ["timed_file", "console"],
            "level": "DEBUG",
            "propagate": True
        },
        "django": {
            "handlers": ["timed_file", "console"],
            "level": "INFO",
            "propagate": True
        },
        "django.request": {
            "handlers": ["timed_file", "console"],
            "level": "ERROR",
            "propagate": True
        }
    },
    "formatters": {
        "verbose": {
            "format": "[%(asctime)s] [%(levelname)s] [%(name)s]: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    }
}


# Internationalization
# https://docs.djangoproject.com/en/5.0/topics/i18n/

LANGUAGE_CODE = 'en-gb'

TIME_ZONE = 'Europe/London'

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.0/howto/static-files/
STATIC_ROOT = BASE_DIR / 'staticfiles'
STATIC_URL = '/static/'

# Extra places for collectstatic to find static files.
STATICFILES_DIRS = (
    BASE_DIR / 'apps/static',
)

# Media Files
MEDIA_ROOT = BASE_DIR / 'media'
MEDIA_URL = '/media/'


# Django Rest Framework
# https://www.django-rest-framework.org/

REST_FRAMEWORK = {
    'DEFAULT_THROTTLE_CLASSES': [
        'rest_framework.throttling.AnonRateThrottle',
        'rest_framework.throttling.UserRateThrottle'
    ],
    'DEFAULT_THROTTLE_RATES': {
        'anon': '100/day',
        'user': '10000/hour'
    },
    "EXCEPTION_HANDLER": "apps.api.exceptions.conflict_exception_handler"
}
